﻿using System;
using System.IO;

namespace Siberix.ReportWriter.Extensions
{
	public class TextOnCircle : Siberix.Report.IExtension
	{
		private string _factoryId;
		private string _id;
		private string _text;
		private Siberix.Graphics.Font _font;
		private Siberix.Graphics.Color _color;
		private float _radius;
		private float _angle;

		public TextOnCircle(string factoryId, string id)
		{
			_factoryId = factoryId;
			_id = id;
		}

		public string FactoryId
		{
			get
			{
				return _factoryId;
			}
		}

		public string Id
		{
			get
			{
				return _id;
			}
		}

		public string Text
		{
			get
			{
				return _text;
			}
			set
			{
				_text = value;
			}
		}

		public Siberix.Graphics.Font Font
		{
			get
			{
				return _font;
			}
			set
			{
				_font = value;
			}
		}

		public Siberix.Graphics.Color Color
		{
			get
			{
				return _color;
			}
			set
			{
				_color = value;
			}
		}

		public float Radius
		{
			get
			{
				return _radius;
			}
			set
			{
				_radius = value;
			}
		}

		public float Angle
		{
			get
			{
				return _angle;
			}
			set
			{
				_angle = value;
			}
		}

		public Siberix.Graphics.Size OnMeasure(Siberix.Graphics.IGraphics graphics)
		{
			graphics.Font = _font;

			float d = (_radius + graphics.FontMetrics.Ascent) * 2.0F;

			return new Siberix.Graphics.Size(d, d);
		}

		public void OnDraw(Siberix.Graphics.IGraphics graphics, float x, float y, float width, float height)
		{
			graphics.Font = _font;
			graphics.Brush = new Siberix.Graphics.SolidColorBrush(_color);

			float cx = x + width / 2.0F;
			float cy = y + height / 2.0F;

			float radius = _radius + graphics.FontMetrics.Ascent;

			float w = 0;
			float h = graphics.FontMetrics.Height;

			int i = 0;
			int cnt = _text.Length;

			char ch = ' ';

			if (cnt > 0)
			{
				ch = _text[0];

				w = graphics.CharWidth(ch);
			}

			float angle = _angle;

			while (i < cnt)
			{
				float angle2 = (float)((angle / 180.0F) * Math.PI);

				float x1 = cx + radius * (float)Math.Cos(angle2);
				float y1 = cy + radius * (float)Math.Sin(angle2);

				graphics.SaveState();
				graphics.Rotate(x1, y1, 90.0F + angle);
				graphics.DrawString(x1 - w / 2.0F, y1, ch.ToString());
				graphics.RestoreState();

				angle += (float)(Math.Atan(w / 2.0F / radius) / Math.PI * 180.0F);

				i++;

				if (i < cnt)
				{
					ch = _text[i];

					w = graphics.CharWidth(ch);

					angle += (float)(Math.Atan(w / 2.0F / radius) / Math.PI * 180.0F);
				}
			}
		}

		public void Save(Stream stream)
		{
			Write(stream, _text);

			lock (_font.Stream)
			{
				Write(stream, _font.Stream);
			}

			Write(stream, _font.Size);
			Write(stream, _color);
			Write(stream, _radius);
			Write(stream, _angle);
		}

		private void Write(Stream stream, int value)
		{
			byte b0 = (byte)(value >> 24);
			byte b1 = (byte)(value >> 16);
			byte b2 = (byte)(value >> 8);
			byte b3 = (byte)value;

			stream.WriteByte(b0);
			stream.WriteByte(b1);
			stream.WriteByte(b2);
			stream.WriteByte(b3);
		}

		private void Write(Stream stream, float value)
		{
			byte[] buffer = BitConverter.GetBytes(value);

			stream.WriteByte(buffer[0]);
			stream.WriteByte(buffer[1]);
			stream.WriteByte(buffer[2]);
			stream.WriteByte(buffer[3]);
		}

		private void Write(Stream stream, Siberix.Graphics.Color color)
		{
			stream.WriteByte(color.A);
			stream.WriteByte(color.R);
			stream.WriteByte(color.G);
			stream.WriteByte(color.B);
		}

		private void Write(Stream stream, string text)
		{
			Write(stream, text.Length);

			foreach (char ch in text)
			{
				byte hb = (byte)(ch >> 8);
				byte lb = (byte)ch;

				stream.WriteByte(hb);
				stream.WriteByte(lb);
			}
		}

		private void Write(Stream stream, Stream s)
		{
			int cnt = (int)s.Length;

			byte[] buffer = new byte[cnt];

			s.Position = 0;
			s.Read(buffer, 0, cnt);

			Write(stream, cnt);

			stream.Write(buffer, 0, cnt);
		}

		public void Load(Stream stream)
		{
			_text = ReadString(stream);

			Stream s = ReadStream(stream);
			
			float size = ReadFloat(stream);

			_font = new Siberix.Graphics.Font(s, size);

			_color = ReadColor(stream);
			_radius = ReadFloat(stream);
			_angle = ReadFloat(stream);
		}

		private int ReadInt(Stream stream)
		{
			int b0 = stream.ReadByte();
			int b1 = stream.ReadByte();
			int b2 = stream.ReadByte();
			int b3 = stream.ReadByte();

			b0 <<= 24;
			b1 <<= 16;
			b2 <<= 8;

			return (int)(b0 | b1 | b2 | b3);
		}

		private float ReadFloat(Stream stream)
		{
			byte[] buffer = new byte[4];

			buffer[0] = (byte)stream.ReadByte();
			buffer[1] = (byte)stream.ReadByte();
			buffer[2] = (byte)stream.ReadByte();
			buffer[3] = (byte)stream.ReadByte();

			return BitConverter.ToSingle(buffer, 0);
		}

		private Siberix.Graphics.Color ReadColor(Stream stream)
		{
			byte a = (byte)stream.ReadByte();
			byte r = (byte)stream.ReadByte();
			byte g = (byte)stream.ReadByte();
			byte b = (byte)stream.ReadByte();

			return new Siberix.Graphics.Color(a, r, g, b);
		}

		private string ReadString(Stream stream)
		{
			int i = 0;
			int cnt = ReadInt(stream);

			System.Text.StringBuilder sb = new System.Text.StringBuilder();

			while (i < cnt)
			{
				int hb = stream.ReadByte();
				int lb = stream.ReadByte();

				sb.Append((char)(hb << 8 | lb));

				i++;
			}

			return sb.ToString();
		}

		private Stream ReadStream(Stream stream)
		{
			int cnt = ReadInt(stream);

			byte[] buffer = new byte[cnt];

			stream.Read(buffer, 0, cnt);

			return new MemoryStream(buffer);
		}
	}
}